/* patricia.c
 * Copyright 2007 (c) Adrian Sai-wah Tam
 * Tue Apr 17 14:13:24 HKT 2007
 *
 * Implementation of Patricia tree in C
 *
 * The idea of Patricia tree implemented here was from:
 *    Donald R. Morrison. "PATRICIA -- practical algorithm to retrieve
 *    information coded in alphanumeric". Journal of the ACM, 15(4):514-534,
 *    October 1968
 * This implementation is standalone in the sense that dependence to external
 * libraries are kept at minimal
 *
 * vim:set ts=8
 */

#include "patricia.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Create a new Patricia tree node with initialization */
patricia_node* patricia_new(char* str, void* data)
{
	patricia_node* N = (patricia_node*)malloc(sizeof(patricia_node));
	N->str = str;
	N->len = 0;
	N->data = data;
	N->left = N->right = NULL;
	return N;
};

/* Copy the content of a Patricia tree node to another */
void patricia_copy(patricia_node* a,const patricia_node* b)
{
	a->str = b->str;
	a->len = b->len;
	a->data = b->data;
	a->left = b->left;
	a->right = b->right;
};

/* Search for a string in the Patricia tree and return the maximal match */
patricia_node* patricia_maxmatch(patricia_node* tree, char* str)
{
	patricia_node* last = NULL;
	while(tree) {
		if ( patricia_bitncmp(str,tree->str,tree->len) ) {
			/* First unmatching node: str != tree->str */
			return last;
		} else if ( patricia_strlen(str) == tree->len ) {
			/* Exactly-matching node: str == tree->str exactly */
			return tree;
		} else {
			/* Partially-matching: tree->str is substring of str */
			if (patricia_strlen(tree->str) == tree->len) {
				last = tree;	/* A terminal node is found */
			};
			tree = patricia_extractbit(str,(tree->len)+1)? tree->left : tree->right;
		};
	};
	/* We went to a nonexisting branch, so the last node is the one we want */
	return last;
};

/* Search for a string in the Patricia tree and return the closest node */
patricia_node* patricia_search(patricia_node* tree, char* str)
{
	patricia_node* last = NULL;
	while(tree) {
		if ( patricia_bitncmp(str,tree->str,tree->len) ) {
			/* First unmatching node: str != tree->str */
			return last;
		} else if ( patricia_strlen(str) == tree->len ) {
			/* Exactly-matching node: str == tree->str exactly */
			return tree;
		} else {
			/* Partially-matching: tree->str is substring of str */
			last = tree;
			tree = patricia_extractbit(str,(tree->len)+1)? tree->left : tree->right;
		};
	};
	/* We went to a nonexisting branch, so the last node is the one we want */
	return last;
};

/* Given a Patricia tree, put the data block into the tree with the specified string as the key */
void patricia_add(patricia_node** tree, char* str, void* data)
{
	patricia_node* node;
	patricia_node* branch;
	patricia_node* cursor;

	/* In case of empty tree: Create and finish */
	if (*tree==NULL) {
		*tree = patricia_new(str,data);
		(*tree)->len = patricia_strlen(str);
		return;
	};
	/* In case we need to change the root node for addition */
	if (patricia_bitncmp((*tree)->str,str,(*tree)->len)) {
		if (patricia_bitncmp((*tree)->str,str,patricia_strlen(str))) {
			/* New node is not a substring of the current root node */
			node = patricia_new(str,NULL);			/* create new root node */
			node->len = patricia_getmatchlen(str,(*tree)->str);
			if (patricia_extractbit(str,(node->len)+1)) {	/* put in two children to the root */
				node->left = patricia_new(str,data);
				node->left->len = patricia_strlen(str);
				node->right = *tree;
			} else {
				node->right = patricia_new(str,data);
				node->right->len = patricia_strlen(str);
				node->left = *tree;
			};
		} else {
			/* New node is a substring of the current root node */
			node = patricia_new(str,data);			/* create new root node */
			node->len = patricia_strlen(str);
			if (patricia_extractbit((*tree)->str,(node->len)+1)) {	/* put in two children to the root */
				node->left = *tree;
			} else {
				node->right = *tree;
			};
		};
		*tree = node;	/* save the root and finish */
		return;
	};
	/* Otherwise: addition is inside the tree */
	cursor = patricia_search(*tree,str);		/* Get the closest matching node, which makes
							   patricia_bitcmp(cursor->str,str,cursor->len)==0 */
	/* In case a duplicated key is found */
	if (patricia_strcmp(cursor->str,str)==0) {
		if (cursor->data != data) {		/* update if data are unequal */
			if (cursor->data) free(cursor->data);
			cursor->data = data;
		};
		return;
	};
	/* In case the node is currently a perfect match but non-terminating */
	if (patricia_strlen(str) == cursor->len) {
		cursor->str = str;			/* make it a terminating node */
		cursor->data = data;
		return;
	};
	/* A new key is being added: Create it first */
	node = patricia_new(str,data);
	node->len = patricia_strlen(str);
	/* In case I shall become a leave node */
	if (cursor->left==NULL && cursor->right==NULL) {
		if (patricia_extractbit(str,(cursor->len)+1)) {
			cursor->left = node;
		} else {
			cursor->right = node;
		};
		return;
	};
	/* Check if I am a substring of an existing node */
	branch = (patricia_extractbit(str,(cursor->len)+1))? (cursor->left) : (cursor->right);
	if (patricia_bitncmp(str,branch->str,patricia_strlen(str))==0) {
		/* Set the branch as my child */
		if (patricia_extractbit(branch->str,(node->len)+1)) {
			node->left = branch;
		} else {
			node->right = branch;
		};
		/* Position myself at the old position of branch */
		if (patricia_extractbit(str,(cursor->len)+1)) {
			cursor->left = node;
		} else {
			cursor->right = node;
		};
		return;
	};
	/* In case of not a substring, a forking node has to be created */
	node->len = patricia_getmatchlen(str,branch->str);	/* Forking node: Length is the bits in common */
	if (patricia_extractbit(str,(cursor->len)+1)) {		/* Replace the old branch with the new fork */
		cursor->left = node;
	} else {
		cursor->right = node;
	};
	cursor = node;
	node = patricia_new(str,data);				/* Create data node */
	node->len = patricia_strlen(str);
	if (patricia_extractbit(str,(cursor->len)+1)) {		/* Add children to the forking node */
		cursor->left = node;
		cursor->right = branch;
	} else {
		cursor->right = node;
		cursor->left = branch;
	};
};

/* Given a Patricia tree, free the key string node->str */
void patricia_freekey(patricia_node* tree, patricia_node* node)
{
	if (tree==NULL || node==NULL) return;
	if (tree == node) {
		free(node->str);
		return;
	};
	if (tree->str == node->str) {	/* Replace every reference to the string */
		tree->str = (tree->str!=tree->left->str) ? (tree->left->str) : (tree->right->str);
	};
	tree = (patricia_extractbit(node->str,(tree->len)+1)) ? (tree->left) : (tree->right);
	patricia_freekey(tree,node);
};

/* Given a Patricia tree, remove the node as indicated by the key */
void patricia_del(patricia_node** tree, char* str)
{
	patricia_node* papa = NULL;
	patricia_node* sibling = NULL;
	patricia_node* cursor = *tree;
	while(cursor) {
		if ( patricia_bitncmp(str,cursor->str,cursor->len) ) {
			return;		/* The key str doesn't exist */
		} else if ( patricia_strlen(str) != cursor->len ) {
			/* cursor->str is a substring of str on first cursor->len bits */
			papa = cursor;
			cursor = patricia_extractbit(str,(cursor->len)+1)? cursor->left : cursor->right;
		} else {	/* key found at cursor */
			if (cursor->left==NULL && cursor->right==NULL) {
				/* If I have no children, clear myself and return */
				if (papa==NULL) {
					/* I am root node */
					*tree = NULL;
				} else if (patricia_strlen(papa->str) == papa->len) {
					/* My parent is a terminating node, unlink myself from my parent */
					if (papa->left == cursor) papa->left = NULL;
					else papa->right = NULL;
				} else {
					/* My parent is not a terminating node, replace my parent with sibling */
					sibling = (papa->left==cursor)? (papa->right) : (papa->left);
					patricia_copy(papa,sibling);
					free(sibling);
				};
				if (cursor->data) free(cursor->data);	/* Then, clear myself */
				patricia_freekey(*tree,cursor);
				free(cursor);
			} else if (cursor->left==NULL || cursor->right==NULL) {
				/* If I get only one children, lift it up to my position */
				if (papa==NULL) {	/* I am root */
					*tree = (cursor->left)? (cursor->left) : (cursor->right);
					if (cursor->data) free(cursor->data);
					patricia_freekey(*tree,cursor);
					free(cursor);
				} else {		/* I am not root */
					sibling = (cursor->left==NULL)? (cursor->right) : (cursor->left);	/* sibling means child here */
					if (cursor->data) free(cursor->data);
					patricia_freekey(*tree,cursor);
					patricia_copy(cursor,sibling);	/* Replace myself with my child */
					free(sibling);
				};
			} else {
				/* I have two children, make myself a non-terminating node */
				if (cursor->data) free(cursor->data);
				patricia_freekey(*tree,cursor);
				cursor->str = cursor->left->str;
			};
			return;
		};
	};
	return;	/* The key doesn't exist */
};

/* ------------------------------------------------------------------------ */
#ifdef UNITTEST

/* Check if a tree is well-formed */
int patricia_validate(patricia_node* tree)
{
	if (tree==NULL) return 0;
	if (tree->left) {
		if (patricia_bitncmp(tree->str,tree->left->str,tree->len)) {
			return 1;	/* Left child is not a substring of myself */
		} else if (patricia_extractbit(tree->left->str,(tree->len)+1)==0) {
			return 1;	/* Right (zero) child located at left */
		} else if (patricia_validate(tree->left)) {
			return 1;	/* Check recurrsively on left branch */
		};
	};
	if (tree->right) {
		if (patricia_bitncmp(tree->str,tree->right->str,tree->len)) {
			return 1;	/* Right child is not a substring of myself */
		} else if (patricia_extractbit(tree->right->str,(tree->len)+1)) {
			return 1;	/* Left (one) child located at right */
		} else if (patricia_validate(tree->right)) {
			return 1;	/* Check recurrsively on right branch */
		};
	};
	return 0;
};

/* Walk through the Patricia tree, and print in braces */
void patricia_printcompact(patricia_node* tree)
{
	if (tree==NULL) return;
	if (patricia_strlen(tree->str) == tree->len) {
		printf("%s{",tree->str);
	} else {
		printf("(%d){",tree->len);
	};
	patricia_printcompact(tree->right);
	printf("}{");
	patricia_printcompact(tree->left);
	printf("}");
};

/* Walk through the Patricia tree, and print in a tree form */
void patricia_print(patricia_node* tree,int n)
{
	int i;
	if (tree==NULL) return;
	for(i=0;i<n;i++) { printf(" "); };
	if (patricia_strlen(tree->str) == tree->len) {
		printf("%s\n",tree->str);
	} else {
		printf("(%d)\n",tree->len);
	};
	patricia_print(tree->right,n+4);
	patricia_print(tree->left,n+4);
};

int main()
{
	char* s;
	int i;
	patricia_node* tree=NULL;
	assert(   0==patricia_bitncmp("\xFF\xFF","\xFF\xFF",16));
	assert(   2==patricia_bitncmp("\xFF\xFF","\xFF\xFD",16));
	assert(   0==patricia_bitncmp("\xFF\xFF","\xFF\xFE",15));
	assert(   0==patricia_bitncmp("\xFF\xFF","\xFF\xF0",12));
	assert(  12==patricia_bitncmp("\xFF\xFF","\xFF\xF1",14));
	assert(0x80==patricia_extractbit("\xA5\xA5",1));
	assert(0x00==patricia_extractbit("\xA5\xA5",2));
	assert(0x01==patricia_extractbit("\xA5\xA5",16));
	assert(0x00==patricia_extractbit("\xA5\xA5",15));
	assert(   0==patricia_getmatchlen("\x01\x01","\x81\x01"));
	assert(   1==patricia_getmatchlen("\x01\x01","\x71\x01"));
	assert(   4==patricia_getmatchlen("\x01\x01","\x09\x01"));
	assert(   8==patricia_getmatchlen("\x01\x01","\x01\xFF"));
	assert(  14==patricia_getmatchlen("\x01\x01","\x01\x02"));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"america");     patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"asia");        patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"asiapacific"); patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"europe");      patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"european");    patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"africa");      patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"australia");   patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"europa");      patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	s = (char*)malloc(sizeof(char)*12); strcpy(s,"antartica");   patricia_add(&tree,s,NULL); assert(0==patricia_validate(tree));
	assert(0==strcmp("asia",patricia_maxmatch(tree,"asian")->str));
	patricia_del(&tree,"europa");      assert(0==patricia_validate(tree));
	patricia_del(&tree,"africa");      assert(0==patricia_validate(tree));
	patricia_del(&tree,"asian");       assert(0==patricia_validate(tree));
	patricia_del(&tree,"asia");        assert(0==patricia_validate(tree));
	patricia_del(&tree,"asiapacific"); assert(0==patricia_validate(tree));
	patricia_del(&tree,"america");     assert(0==patricia_validate(tree));
	patricia_del(&tree,"europe");      assert(0==patricia_validate(tree));
	patricia_del(&tree,"australia");   assert(0==patricia_validate(tree));
	patricia_del(&tree,"antartica");   assert(0==patricia_validate(tree));
	patricia_del(&tree,"european");    assert(0==patricia_validate(tree));
	printf("OK\n");
	for (i=0; i<1e5; i++) {
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"america");     patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"asia");        patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"asiapacific"); patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"europe");      patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"european");    patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"africa");      patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"australia");   patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"europa");      patricia_add(&tree,s,NULL); patricia_validate(tree);
		s = (char*)malloc(sizeof(char)*12); strcpy(s,"antartica");   patricia_add(&tree,s,NULL); patricia_validate(tree);
		patricia_maxmatch(tree,"asian");
		patricia_del(&tree,"europa");      patricia_validate(tree);
		patricia_del(&tree,"africa");      patricia_validate(tree);
		patricia_del(&tree,"asian");       patricia_validate(tree);
		patricia_del(&tree,"asia");        patricia_validate(tree);
		patricia_del(&tree,"asiapacific"); patricia_validate(tree);
		patricia_del(&tree,"america");     patricia_validate(tree);
		patricia_del(&tree,"europe");      patricia_validate(tree);
		patricia_del(&tree,"australia");   patricia_validate(tree);
		patricia_del(&tree,"antartica");   patricia_validate(tree);
		patricia_del(&tree,"european");    patricia_validate(tree);
	};
	return 0;
};
#endif

#ifdef __cplusplus
}
#endif
